/*******************************************************************************
 mydclgen copyright 1998 by Peter Hellmann  -  all rights reserved 

 author:  Peter Hellmann
 version: see symbolic constant MY_VERSION
 e-mail:  hellmann@w-4.de
 website: http://www.w-4.de/~hellmann
 
*******************************************************************************/ 


#include <stdio.h>
#include <string.h>
#include <memory.h>


#define MAXDBNAMELEN     32
#define MAXWRITEBUFLEN  500
#define SQL_NOT_FOUND   100
#define MY_VERSION     "1.4"

char* StrTrim( char* str );
void StrToUcase(char* str);
void MapDbDatatypeC(char* col, char* dbt, int len, char* mbr);
void DisplaySyntax(char* pn);

EXEC SQL INCLUDE sqlca;

EXEC SQL BEGIN DECLARE SECTION;
  long  rowcount;
  char  dbname[33];
  int   column_sequence;
  int   column_length;
  long  column_amount;
  char  column_name[33];
  char  column_datatype[33];
  char  table_name[33];
  char * stmtSelect1 = "SELECT table_name FROM iitables WHERE table_owner = 'autefa' AND table_type = 'T' ORDER BY table_name; ";
  char * stmtSelect2 = "SELECT column_name,column_datatype,column_length,column_sequence FROM iicolumns WHERE table_name = ? ORDER BY column_sequence; ";
  char * stmtSelect3 = "SELECT iitables.table_name,iicolumns.column_name,iicolumns.column_datatype,iicolumns.column_sequence,iicolumns.column_length"
                       " FROM iitables,iicolumns WHERE iitables.table_owner = 'autefa' AND iitables.table_type = 'T'"
                       " AND (iicolumns.column_datatype = 'CHAR' OR iicolumns.column_datatype = 'VARCHAR') AND iicolumns.table_name = iitables.table_name ORDER BY iitables.table_name,iicolumns.column_sequence; ";
  char * stmtSelect4 = "SELECT iitables.table_name,iicolumns.column_name,iicolumns.column_sequence "
                       " FROM iitables,iicolumns WHERE iitables.table_owner = 'autefa' AND iitables.table_type = 'T'"
                       " AND iicolumns.table_name = iitables.table_name ORDER BY iitables.table_name,iicolumns.column_sequence; ";
  char * stmtSelect5 = "SELECT COUNT(*) FROM iicolumns WHERE table_name = ?; ";
EXEC SQL END DECLARE SECTION;

/*******************************************************************************
 main function
*******************************************************************************/ 

int main(int argc, char *argv[])
{
char  the_define[100];
char  the_member[100];
char  the_dim[100];
  
long i;
FILE* fp;
char* p;
char blank[] = "                                                            ";
char writebuf[MAXWRITEBUFLEN + 1]; 
char col_len[20];

// display command line arguments
//for (i=0;i<argc;i++)
//    {
//    printf("argc:%d argv:%s\n", i, argv[i]);
//    }

// check for all required command line arguments
if (argc != 4)
   {
   DisplaySyntax(argv[0]);
   exit(0);
   }

// check string length of databasename
if (strlen(argv[1]) > MAXDBNAMELEN)
   {
   printf("databasename too long, only %d characters allowed !\n\n", MAXDBNAMELEN);
   exit(0);
   }
else
   {
   // copy argument to host variable
   strcpy(dbname, argv[1]);
   }

// connect to database
EXEC SQL CONNECT :dbname;
if (sqlca.sqlcode < 0)
   {
   printf("could not connect to database '%s' !!!\n", dbname);
   exit(0);
   }
else
   {   
   printf("successfully connected to database '%s'\n", dbname);
   }

// create output file
fp = fopen(argv[3], "w");

if(!fp)
  {
  printf("sorry, could not create output file '%s' !\n\n", argv[3]);
  exit(0);
  }
else
  {
  printf("output file '%s' created successfully\n", argv[3]);
  }  
  
printf("working on job...\n");  

// print a header ...
sprintf(writebuf,"/*****************************************************************\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  program:   mydclgen for C\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  copyright: 1998 by Peter Hellmann  -  all rights reserved -\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  author:    Peter Hellmann\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  version:   %s\r\n", MY_VERSION); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  e-mail:    hellmann@w-4.de\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  website:   http://www.w-4.de/~hellmann\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"******************************************************************/\r\n\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);

sprintf(writebuf,"/******************************************************************\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"*  database:  %s\r\n", dbname); fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf,"******************************************************************/\r\n\r\n"); fwrite(writebuf, strlen(writebuf), 1, fp);
// query database

EXEC SQL PREPARE stmtSelect1 FROM :stmtSelect1;
EXEC SQL DECLARE mycursor1 CURSOR FOR stmtSelect1;

EXEC SQL PREPARE stmtSelect2 FROM :stmtSelect2;
EXEC SQL DECLARE mycursor2 CURSOR FOR stmtSelect2;

EXEC SQL PREPARE stmtSelect3 FROM :stmtSelect3;
EXEC SQL DECLARE mycursor3 CURSOR FOR stmtSelect3;

EXEC SQL PREPARE stmtSelect4 FROM :stmtSelect4;
EXEC SQL DECLARE mycursor4 CURSOR FOR stmtSelect4;

EXEC SQL PREPARE stmtSelect5 FROM :stmtSelect5;
EXEC SQL DECLARE mycursor5 CURSOR FOR stmtSelect5;

// check what job we have to do
if (strcmp(argv[2],"struct") == 0) goto MakeAStructFile;  
if (strcmp(argv[2],"const") == 0) goto MakeAConstFile;

// user didn't dig syntax...
DisplaySyntax(argv[0]);
exit(0);


MakeAConstFile:

// generate symbolic constants for string length of columns

// fill write buffer
sprintf(writebuf, "\r\n/* stringlengths (including zero termination byte) for CHAR() columns:*/\r\n\r\n");
// write to output file
fwrite(writebuf, strlen(writebuf), 1, fp);

printf("generating stringlengths\n");

EXEC SQL OPEN mycursor3;

while (1)
      {
      EXEC SQL FETCH mycursor3 INTO :table_name, :column_name, :column_datatype, :column_sequence, :column_length;
      if (sqlca.sqlcode == SQL_NOT_FOUND || sqlca.sqlcode < 0) break;
      // fill write buffer
      StrTrim(table_name);
      strcat(table_name, "\0");
      StrTrim(column_name);
      strcat(column_name, "\0");
      StrToUcase(table_name);
      memset(the_define,0x20,strlen(the_define));
      sprintf(the_define, "#define %s_%s_LEN\0", table_name, column_name);
      strncat(the_define, blank, 70-strlen(the_define));
      sprintf(writebuf, "%s %d\r\n\0", the_define, column_length + 1);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      }

EXEC SQL CLOSE mycursor3;     

printf("generating stringlengths done\n");

// generate symbolic constants for sequence number of columns

// fill write buffer
sprintf(writebuf, "\r\n/* column-index (for null value array):*/\r\n\r\n");
// write to output file
fwrite(writebuf, strlen(writebuf), 1, fp);

printf("generating column-indices\n");

EXEC SQL OPEN mycursor4;

while (1)
      {
      EXEC SQL FETCH mycursor4 INTO :table_name, :column_name, :column_sequence;
      if (sqlca.sqlcode == SQL_NOT_FOUND || sqlca.sqlcode < 0) break;
      // fill write buffer
      StrTrim(table_name);
      strcat(table_name, "\0");
      StrTrim(column_name);
      strcat(column_name, "\0");
      StrToUcase(table_name);
      memset(the_define,0x20,strlen(the_define));
      sprintf(the_define, "#define %s_%s_INX\0", table_name, column_name);
      strncat(the_define, blank, 70-strlen(the_define));
      sprintf(writebuf, "%s %d\r\n\0", the_define, column_sequence - 1);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      }

EXEC SQL CLOSE mycursor4;  

printf("generating column-indices done\n");

// generate null value dim constants

// fill write buffer
sprintf(writebuf, "\r\n/* null-value array dimensions:*/\r\n\r\n");
// write to output file
fwrite(writebuf, strlen(writebuf), 1, fp);

printf("generating null-value array dimensions\n");

EXEC SQL OPEN mycursor1;

while (1)
      {
      EXEC SQL FETCH mycursor1 INTO :table_name;
      if (sqlca.sqlcode == SQL_NOT_FOUND || sqlca.sqlcode < 0) break;
      EXEC SQL OPEN mycursor5 USING :table_name;
      EXEC SQL FETCH mycursor5 INTO :column_amount;
      EXEC SQL CLOSE mycursor5;     
      StrTrim(table_name);
      strcat(table_name, "\0");
      memset(the_dim,0x20,strlen(the_dim));
      sprintf(the_dim, "#define %s_DIM\0", table_name);
      strncat(the_dim, blank, 50-strlen(the_dim));
      // fill write buffer
      sprintf(writebuf, "%s %d\r\n", the_dim, column_amount);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      }
      
EXEC SQL CLOSE mycursor1;   

printf("generating null-value array dimensions done\n");

// fill write buffer
sprintf(writebuf, "\r\n/* NULL value symbolic constants (for null value array):*/\r\n\r\n");
// write to output file
fwrite(writebuf, strlen(writebuf), 1, fp);

sprintf(writebuf, "#define SQL_INDICATOR_NULL                        -1\r\n");
fwrite(writebuf, strlen(writebuf), 1, fp);
sprintf(writebuf, "#define SQL_INDICATOR_NOT_NULL                     0\r\n");
fwrite(writebuf, strlen(writebuf), 1, fp);

goto Finish;

MakeAStructFile:

// generate structs for embedded sql using C

// fill write buffer
sprintf(writebuf, "\r\n/* C structs:*/\r\n\r\n");
// write to output file
fwrite(writebuf, strlen(writebuf), 1, fp);

printf("generating C structs");

EXEC SQL OPEN mycursor1;

while (1)
      {
      EXEC SQL FETCH mycursor1 INTO :table_name;
      if (sqlca.sqlcode == SQL_NOT_FOUND || sqlca.sqlcode < 0) break;

      StrTrim(table_name);
      strcat(table_name, "\0");
      
      // fill write buffer
      sprintf(writebuf, "/* struct buffer for table %s:*/\r\n\r\n", table_name);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      
      // fill write buffer
      sprintf(writebuf, "struct S%s_\r\n{\r\n", table_name);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      EXEC SQL OPEN mycursor2 USING :table_name;
      while (1)
            {
            EXEC SQL FETCH mycursor2 INTO :column_name, :column_datatype, :column_length, :column_sequence;
            if (sqlca.sqlcode == SQL_NOT_FOUND || sqlca.sqlcode < 0) break;
            
            memset(the_member,0x20,strlen(the_member));
            StrTrim(column_name);
            strcat(column_name, "\0");
            StrTrim(column_datatype);
            strcat(column_datatype, "\0");
            
            MapDbDatatypeC(column_name, column_datatype, column_length, the_member);
            
            // fill write buffer
            sprintf(writebuf, "  %s\r\n", the_member);
            // write to output file
            fwrite(writebuf, strlen(writebuf), 1, fp);
            }

      EXEC SQL CLOSE mycursor2; 
      sprintf(writebuf, "} S%s;\r\n\r\n", table_name);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      
      // fill write buffer
      sprintf(writebuf, "/* null value array for table %s:*/\r\n\r\n", table_name);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      
      // generate the null value array for the table
      EXEC SQL OPEN mycursor5 USING :table_name;
      EXEC SQL FETCH mycursor5 INTO :column_amount;
      EXEC SQL CLOSE mycursor5;     

      StrTrim(table_name);
      strcat(table_name, "\0");

      // fill write buffer
      sprintf(writebuf, "short N%s[%ld];\r\n\r\n", table_name, column_amount);
      // write to output file
      fwrite(writebuf, strlen(writebuf), 1, fp);
      
      }
      
EXEC SQL CLOSE mycursor1;     

printf("\ngenerating C structs done\n");

Finish:

// close output file
fclose(fp);

printf("job done\n");

// exit program
exit(0);
}


char* StrTrim(char* str)
{
char *p = str;
while( *p ) p++;
p--;
while( *p == ' ' || *p == '\t' ) p--;
*++p = '\0';
return str;
};

void StrToUcase(char* str)
{
char *p = str;
while( *p )
     {
     *p = toupper(*p);
     p++;
     }
};


void MapDbDatatypeC(char* col, char* dbt, int len, char* mbr)
{

sprintf(mbr,"%s %s", dbt, col);

// map text types
if(strcmp(dbt,"CHAR") == 0) { sprintf(mbr,"char %s[%d];", col, len + 1); return; }
if(strcmp(dbt,"VARCHAR") == 0) { sprintf(mbr,"char %s[%d];", col, len + 1); return; }
if(strcmp(dbt,"TEXT") == 0) { sprintf(mbr,"char %s[%d];", col, len + 1); return; }

// map integer types
if(strcmp(dbt,"INTEGER") == 0)
  {
  if (len >= 4)
     {
     sprintf(mbr,"int %s;", col);
     }
  else
     {
     sprintf(mbr,"short %s;", col);
     }   
  return;
  }

// map floating point types
if(strcmp(dbt,"DECIMAL") == 0) { sprintf(mbr,"double %s;", col); return; }
if(strcmp(dbt,"FLOAT") == 0)
  {
  if (len >= 8)
     {
     sprintf(mbr,"double %s;", col);
     }
  else
     {
     sprintf(mbr,"float %s;", col);
     }   
  return;
  }
  
// misc types  
if(strcmp(dbt,"MONEY") == 0) { sprintf(mbr,"double %s;", col); return; }
if(strcmp(dbt,"DATE") == 0) { sprintf(mbr,"char %s[26];", col); return; }
if(strcmp(dbt,"BYTE") == 0)
  {
  sprintf(mbr,"struct %s_\r\n\t{\r\n\tshort\tlen;\r\n\tchar\tbuf[%d];\r\n\t} %s;", col, len, col);
  return;
  }

sprintf(mbr,"unknown %s", col);
printf("\nWARNING !!! unknown datatype: %s\n", dbt);
return;
};


void DisplaySyntax(char* pn)
{
   printf("\n***************************************************************\n");
   printf("program:   mydclgen for C\n");
   printf("copyright: 1998 by Peter Hellmann  -  all rights reserved -\n");
   printf("version:   %s\n", MY_VERSION);   
   printf("author:    Peter Hellmann\n");
   printf("e-mail:    hellmann@w-4.de\n");
   printf("website:   http://www.w-4.de/~hellmann\n");
   printf("***************************************************************\n\n");
   printf("this tiny program creates automatically C structs for all tables\n");
   printf("to use with embedded SQL in C. NULL value arrays are created too.\n");   
   printf("you can also create automatically C symbolic constants that hold\n");
   printf("lengths for strings respectively indices to use with NULL value arrays\n");   
   printf("\nplease use the following syntax:\n");
   printf("%s <databasename> <struct/const> <outputfilename>\n\n", pn);
   printf("example for generation of structs:\n");
   printf("%s mycooldatabase struct myCoolStructs.h\n\n", pn);
   printf("example for generation of constants:\n");
   printf("%s mycooldatabase const myCoolConstants.h\n", pn);
   printf("***************************************************************\n\n");   
}








